;asm6502 asm\pcm-high.asm -P32 -M0 -Npcm-high.nes -L
;
; nespcm.asm
;  No interpolation, high-sample rate output, as-is sample rates.
;
;  This was accomplished by adding full fixed-point rates, and tightening the
; loop as much as possible, plus one cycle for an output rate of 33,143.93939Hz,
; the same as the DMC channel for the highest bitrate, but full PCM samples not
; deltas.
;  Rates above the output rate should have excess nulls, as the sample reader
; reads in an aliased fashion.  (May skip over one sample for every time the
; rate divides by 33,143.93939.)

samplePtr	= 0 ;fixed-point 16:16
sampleRate	= 4 ;fixed-point 23:1

	.opt	on

ORG	$8000
	;540672/546875 ~= 34603/35000 ~= 31637/32000
snd0	dc.w	(22050*31637+16000)/32000
	.incbin	"loon-2.raw"
	dc.b	0
snd1	dc.w	(11127*31637+16000)/32000
	.incbin	"tada.raw"
	dc.b	0
snd2	dc.w	(22255*31637+16000)/32000
	.incbin	"indigo.raw"
	dc.b	0
	;managed to squeeze in three waves - 6/16, 4/16, and 2/16 amplitude
snd3	dc.b	$A0,$C0,$E0,$E0,$E0,$E0,$E0,$E0,$C0,$A0,$80,$80,$80,$80,$80,$80
	dc.b	0
snd4	dc.b	$95,$AA,$C0,$C0,$C0,$C0,$C0,$C0,$AA,$95,$80,$80,$80,$80,$80,$80
	dc.b	0
snd5	dc.b	$8A,$95,$A0,$A0,$A0,$A0,$A0,$A0,$95,$8A,$80,$80,$80,$80,$80,$80
	dc.b	0

playSound
	ldy	#0
	sta	$2000	; best to turn NMIs off during play
	sty	samplePtr
	sty	samplePtr+1
	sta	samplePtr+2
	stx	samplePtr+3
	sty	sampleRate+2

	; load the sample rate
	lda	(samplePtr+2),y
	sta	sampleRate
	inc	samplePtr+2
	bne	+

	inc	samplePtr+3
+	lda	(samplePtr+2),y
	sta	sampleRate+1
	inc	samplePtr+2
	bne	+

	inc	samplePtr+3
	ora	sampleRate
+	beq	_pse	;zero sample rate?!

	asl	sampleRate
	rol	sampleRate+1
	rol	sampleRate+2
_psr	ldy	#$10
	sty	$4015
_pse	rts

playSndRate
	sta	samplePtr+2
	stx	samplePtr+3
	jmp	_psr

reset:	sei

	lda	#$40	; disable APU frame IRQ
	sta	$4017
	lda	$4015	; clear any already-pending IRQ

	lda	#<(filter>>6)
	sta	$4012	; DMC start addr
	lda	#0	; DMC len = 1 sample
	sta	$4013
	lda	#$8F	; initial values for IRQ handler
	sta	$4010

	lda	#$8D
	sta	$2001

	cli		; enable interrupts
replay
	lda	#<snd0
	ldx	#>snd0
	jsr	playSound

	lda	#<snd1
	ldx	#>snd1
	jsr	playSound

	lda	#<snd2
	ldx	#>snd2
	jsr	playSound

	ldy	#<(920*16)
	sty	sampleRate
	ldy	#>(920*16)
	sty	sampleRate+1
	ldy	#0
	sty	sampleRate+2
	ldx	#>snd3
	lda	#<snd3
	ldy	#40
	sty	$10
-	jsr	playSndRate
	dec	$10
	bne	-

	ldx	#>snd4
	lda	#<snd4
	ldy	#20
	sty	$10
-	jsr	playSndRate
	dec	$10
	bne	-

	ldx	#>snd5
	lda	#<snd5
	ldy	#10
	sty	$10
-	jsr	playSndRate
	dec	$10
	bne	-
	jmp	replay

; NMI handler
nmi:	rti

ORG	$FE80
filter	dc.b    $AA,0,1,1,1,2,2,2,3,3,3,4,4,4,5,5
	dc.b    5,6,6,6,7,7,7,8,8,8,9,9,10,10,10,11
	dc.b    11,11,12,12,12,13,13,14,14,14,15,15,15,16,16,16
	dc.b    17,17,18,18,18,19,19,20,20,20,21,21,21,22,22,23
	dc.b    23,23,24,24,25,25,25,26,26,27,27,27,28,28,29,29
	dc.b    29,30,30,31,31,32,32,32,33,33,34,34,34,35,35,36
	dc.b    36,37,37,38,38,38,39,39,40,40,41,41,42,42,42,43
	dc.b    43,44,44,45,45,46,46,47,47,47,48,48,49,49,50,50
	dc.b    51,51,52,52,53,53,54,54,55,55,56,56,57,57,58,58
	dc.b    59,59,60,60,61,61,62,62,63,63,64,64,65,65,66,66
	dc.b    67,68,68,69,69,70,70,71,71,72,72,73,74,74,75,75
	dc.b    76,76,77,78,78,79,79,80,80,81,82,82,83,83,84,84
	dc.b    85,86,86,87,88,88,89,89,90,91,91,92,92,93,94,94
	dc.b    95,96,96,97,98,98,99,99,100,101,101,102,103,103,104,105
	dc.b    105,106,107,107,108,109,110,110,111,112,112,113,114,114,115,116
	dc.b    117,117,118,119,119,120,121,122,122,123,124,125,125,126,127,127

irq:	pha
	tya
	pha
	txa
	pha
	clc
	ldy	#0
	jmp	sstart

sloop	tax
	lda	filter,x
	sta	$4011
	lda	samplePtr
	adc	sampleRate
	sta	samplePtr
	lda	samplePtr+1
	adc	sampleRate+1
	sta	samplePtr+1
	lda	samplePtr+2
	adc	sampleRate+2
	sta	samplePtr+2
	lda	samplePtr+3
	adc	#0
	sta	samplePtr+3
sstart	lda	(samplePtr+2),y
	bne	sloop

	lda	filter,y
	sta	$4011
	lda	#0
snde	sta	$4015
	pla
	tax
	pla
	tay
	pla
	rti

ORG	$FFFA

	dc.w	nmi
	dc.w	reset
	dc.w	irq
