;asm6502 asm\pcm-high5.asm -P32 -M5 -L -Npcm-high5.nes
;
; pcm-high5.asm
;  No interpolation, high sample rate output, using MMC5
;
; (C) 2015 Adam J. P.
;
;  With MMC5 PCM, no DMC-filter is necessary; full 8 bit output
; (except for the terminating null).
;
;  The original sound data, I am not responsible for, but is
; publicly available for free in Apple/Software Updates/System 7 and
; HyperCard Player.

;.define doscale
.define dompython

; ==== globals

; X is used for the lowest 8 bits of the 32-bit fixed-point position
samplePtr	= 0	;fixed-point 16:16 (middle bytes not necessarily updated)
sampleRate	= 4	;4 bytes (sometimes only 3 used)
tmp		= 8	;8 bytes
PCMEnable	= 16

waveTab		= $200

; ==== data

	.opt	on

ORG	$8000
 .if 0
snd0	dcw	22050
	dcb	60
	.incbin	"loon-2.raw"
	dcb	0
snd1	dcw	11127
	dcb	60
	.incbin	"tada.raw"
	dcb	0
 .endif
snd2	dcw	22255
	dcb	67
	.incbin	"harpsichord.raw"
	dcb	0

ORG	$F000

; ==== library

; mulu16
;  Uses the MMC5 multiplier to compute a full 16 by 16 multiply.
;
;  Parameters:
;   X:Y and tmp:A
;  Result in tmp (big-endian).  A=copy of high byte.  X:Y maintained.

mulu16
	sta	$5205	;mmc5m0
	sty	$5206	;mmc5m1
	lda	$5205	;mull
	sta	tmp+3
	lda	$5206	;mulh
	sta	tmp+2
	stx	$5206	;mmc5m1
	clc
	lda	tmp+2
	adc	$5205
	sta	tmp+2
	lda	$5206	;mulh
	adc	#0
	sta	tmp+1
	lda	tmp
	sta	$5205	;mmc5m0
	lda	#0
	rol
	sta	tmp
	sty	$5206	;mmc5m1
	lda	tmp+2
	adc	$5205	;mull
	sta	tmp+2
	lda	tmp+1
	adc	$5206	;mulh
	sta	tmp+1
	lda	tmp
	adc	#0
	sta	tmp
	stx	$5206	;mmc5m1
	lda	tmp+1
	adc	$5205	;mull
	sta	tmp+1
	lda	tmp
	adc	$5206	;mulh
	sta	tmp
	rts

; playSound
;  Plays a PCM sampled sound, interpreting a null (#$00) as a terminator.
;
;  Parameters:
;   X:A is the 16-bit address to read from, with sample rate header.
;   Y is the base note; 0=default.

playSound
	sta	samplePtr+2
	stx	samplePtr+3
	sty	sampleRate+2
	lax	#0
	tay
	sty	samplePtr

	; load the sample rate
	lda	(samplePtr+2),y
	iny
	sta	sampleRate
	lda	(samplePtr+2),y
	iny
	sta	sampleRate+1
	; apply base note - X:A
	lda	sampleRate+2
	bpl	+
	; default rate
	lda	#120
	bne	@ssr

+	adc	#120
	sec
	sbc	(samplePtr+2),y
	bcs	+
	rts	; out of range

@ssr	tay
	lda	#$2A	; divide by 12
	sta	tmp
	lda	#$AB
	jsr	mulu16
	lda	tmp+1	; store shift ct, get remainder
	lsr
	sta	$5205	;mmc5m0
	sec
	sbc	#16
	sta	sampleRate+2
	lda	#12
	sta	$5206	;mmc5m1
	tya
	sec
	sbc	$5205	;mull
	asl
	tax
	ldy	noteTab,x
	lda	noteTab+1,x
	tax
	lda	sampleRate+1
	sta	tmp
	lda	sampleRate
	jsr	mulu16
	ldx	sampleRate+2
	bpl	@nosh

-	lsr
	ror	tmp+1
	ror	tmp+2
	inx
	bne	-

@nosh	tax
	lda	#0
	adc	tmp+2	;now copy, rounded
	sta	sampleRate
	lda	tmp+1
	adc	#0
	sta	sampleRate+1
	txa
	adc	#0
	sta	sampleRate+2
	ora	sampleRate+1
	ora	sampleRate
	beq	@pse	;zero sample rate?

	lda	samplePtr+2
	adc	#3
	bcc	+

	inc	samplePtr+3
+	tay
	ldx	#0
	stx	samplePtr+2
	beq	@pss

 ;40+2/256 cc overall (~44,735.580763882Hz)

  ;Hz in twenty-thousandths
;srate		= 441876573*2
srate		= 447355808*2

srcale		= 4 ;highest value possible without overflow in noteTab

@psh	inc	samplePtr+3	;5
	lda	(samplePtr+2),y	;5
	beq	@pse		;3
				;-1=12
@psl	sta	$5011		;4
	lda	samplePtr	;3
	adc	sampleRate	;3
	sta	samplePtr	;3
	txa			;2
	adc	sampleRate+1	;3
	tax			;2
	tya			;2
	adc	sampleRate+2	;3
	tay			;2
	bcs	@psh		;+3=30
				;-1
	lda	PCMEnable	;3
	beq	@pse		;3
				;-1
@pss	lda	(samplePtr+2),y	;5
	beq	@pse		;3
				;-1
	sta	$5011		;4
	lda	samplePtr	;3
	adc	sampleRate	;3
	sta	samplePtr	;3
	txa			;2
	adc	sampleRate+1	;3
	tax			;2
	tya			;2
	adc	sampleRate+2	;3
	tay			;2
	bcs	@psh		;3
				;-1
	nop			;2
	lda	(samplePtr+2),y	;5
	bne	@psl		;3
@pse	rts

 .if 0
; play88 - untested
;  Plays 8-bit PCM sound at approximately 88.2KHz.
;  Can be used to play at high rates or dithered CD rate audio.
;  Zero on an even byte terminates.
; Parameters:
;  X:Y is the address, and must be on a 2-byte boundary.
; Uses: A, X, Y, samplePtr[2] and sampleRate[1].

play88
 ;20+148/514+1/256 cc = ~88,201.58 Hz
	stx	samplePtr+3
	lda	#0
	sta	samplePtr+2
	sta	sampleRate
	beq	@p88s

@p88n	inc	samplePtr+3	;+5=21
@p88w	sta	$5011		;4
@p88s	lda	(samplePtr+2),y	;5
	beq	@p88e		;3
				;-1
	iny			;2
	beq	@p88n		;+3
				;-1
	ldx	PCMEnable	;3
	beq	@p88e		;+3
				;-1=20
	sta	$5011		;4
	lda	sampleRate	;3
	adc	#148		;2
	sta	sampleRate	;3
	bcs	@p88o		;3
				;-1
@p88o	lda	(samplePtr+2),y	;5
	iny			;2 - will not be zero
	bne	@p88w		;3
@p88e	rts
 .endif

 .if 0
; play22050_10 - untested
;  Plays a 16-bit PCM sampled sound with dither encoding at 22050Hz.
;  The format is unsigned, and a 0 sample terminates.
; Parameters:
;  X:Y is the address, and must be on a 2-byte boundary.
; Uses: A, X, Y, samplePtr[2] and sampleRate[1].

play22050_10
 ;81+43/256 cc = ~22050.23428374 Hz, with a 22,051.2955 peak
 ;(needs to be 81.16883116883446712 cc for true 22050)
	stx	samplePtr+3
	lda	#0
	sta	samplePtr+2
	sta	sampleRate
	clc
	beq	@p10s

@p10n	inc	samplePtr+3	;5
	clc			;+2=82
@p10s	lda	(samplePtr+2),y	;5
	sta	samplePtr	;3
	iny			;2
	lda	(samplePtr+2),y	;5
	beq	@p10e		;3
				;-1
	adc	#-1		;+2=19
	tax			;2
	asl	samplePtr	;5
	adc	#0		;2
	sta	$5011		;+4=13(x3)
	txa			;2
	asl	samplePtr	;5
	adc	#0		;2
	sta	$5011		;4
	txa			;2
	asl	samplePtr	;5
	adc	#0		;2
	sta	$5011		;4
	lda	sampleRate	;3
	adc	#43-1		;2
	sta	sampleRate	;3
	iny			;2
	stx	$5011		;+4=14
	beq	@p10n		;+3=75
				;-1
	bcs	@p10c		;3
				;-1
@p10c	nop			;2
	jmp	@p10s		;-1=81+c
@p10e	rts
 .endif

 .if 0
  ; Full linear interpolation, using the MMC5 multiplier.
  ; The highest DMC IRQ-based sample rate (12,428.97727 Hz)
  ;would leave only 10 cycles to spare (2.315% CPU cyles available),
  ;which obviously wouldn't work with action games...
  ; The implementation here runs at double (8,285.98485 Hz),
  ;a.k.a "telephone quality."  1.0 must be subtracted from the rate.
  ; The guts of the interpolation rtn can be optimized down to 98 cc
  ;(99 looped, 18,078.5124 Hz).

;srate		= 82859848*2
;srcale		= 1

irql	lda	tmp		;3
	sta	$5011		;4
	; insert reload function (indirect ptr call) here
irqe0	jmp	irqe		;3

irqp0	sta	$5205		;4 - mmc5m0
	lda	tmp		;3
	adc	$5206		;4 - mulh
	bcc	irqs0		;3

irq:	pha			;3
	tya			;2
	pha			;+3=8
	;0
	ldy	#0		;2
	lda	(samplePtr+2),y	;5
	beq	irqe0		;3
				;-1
	sta	tmp		;3
	lda	samplePtr+2	;3
	clc			;2
	adc	#1		;2
	sta	samplePtr+2	;3
	tya			;2
	adc	samplePtr+3	;3
	sta	samplePtr+3	;3
	lda	(samplePtr+2),y	;5
	beq	irql0		;3
				;-1=37
	ldy	samplePtr+1	;3
	sty	$5206		;4 - mmc5m1
	sec			;2
	sbc	tmp		;3
	bcs	irqp0		;3
				;-1
	adc	#-1		;2
	eor	#-1		;2
	sta	$5205		;4 - mmc5m0
	lda	samplePtr+1	;3
	sbc	$5206		;4 - mulh
irqs0	sta	$5011		;+4=33
	lda	samplePtr	;2
	adc	sampleRate	;3
	sta	samplePtr	;3
	tya			;2
	adc	sampleRate+1	;3
	sta	samplePtr+1	;3
	lda	samplePtr+2	;3
	adc	sampleRate+2	;3
	sta	samplePtr+2	;3
	lda	samplePtr+3	;3
	adc	sampleRate+3	;3
	sta	samplePtr+3	;+3=34
				;104
	;1 - delay so samples are more evenly apart (216 cc)
	; this can be replaced w. a cyc-consistent IRQ routine
	ldy	#14		;2
-	dey			;2
	sty	tmp+1		;3
	bne	-		;3
				-1=113
	lda	(samplePtr+2),y	;5
	beq	irqe		;3
				;-1
	sta	tmp		;3
	lda	samplePtr+2	;3
	clc			;2
	adc	#1		;2
	sta	samplePtr+2	;3
	tya			;2
	adc	samplePtr+3	;3
	sta	samplePtr+3	;3
	lda	(samplePtr+2),y	;5
	beq	irql1		;3
				;-1=37
	ldy	samplePtr+1	;3
	sty	$5206		;4 - mmc5m1
	sec			;2
	sbc	tmp		;3
	bcs	irqp1		;3
				;-1
	adc	#-1		;2
	eor	#-1		;2
	sta	$5205		;4 - mmc5m0
	lda	samplePtr+1	;3
	sbc	$5206		;4 - mulh
irqs1	sta	$5011		;+4=33
	lda	samplePtr	;2
	adc	sampleRate	;3
	sta	samplePtr	;3
	tya			;2
	adc	sampleRate+1	;3
	sta	samplePtr+1	;3
	lda	samplePtr+2	;3
	adc	sampleRate+2	;3
	sta	samplePtr+2	;3
	lda	samplePtr+3	;3
	adc	sampleRate+3	;3
	sta	samplePtr+3	;+3=34
				;104
	ldy	#16		;2
irqe	sty	$4015		;4
	pla			;4
	tay			;2
	pla			;4
	rti			;+6=31
				;+8+216=359 (83.1% CPU when in use)
irql1	lda	tmp		;3
	sta	$5011		;4
	lda	#0
	; insert reload function (indirect ptr call) here
	beq	irqe		;3

irqp1	sta	$5205		;4 - mmc5m0
	lda	tmp		;3
	adc	$5206		;4 - mulh
	bcc	irqs1		;3
 .endif

; playWaveLenCtr
;  Plays the wave on loop, length counter approach.
;  Parameters:
;   X:A is the fixed-point no. of wave cycles to play; 0=256.0.
;   Y is the step delay; 1=the highest freq; 0=256, the longest.
;
;  Uses samplePtr[2] and sampleRate[1] as temporary storage.

 .if 0
playWaveLenCtr
	sta	samplePtr
	stx	samplePtr+2
	sty	sampleRate
	ldx	#0
	beq	@wtbs
	;shortest loop (Y=1) is 27cc
	;makes an output rate of ~66,287.8788Hz (a wave of ~258.9370265Hz)
@wtbl	dec	samplePtr+2	;5
	beq	@wtbe		;3
				;-1+3 = 10
	inx			;2 - X=1
	lda	waveTab+1	;4
	sta	$5011		;4
	ldy	sampleRate	;3
-	dey			;2
	bne	-		;3
				;-1
	dec	samplePtr	;5 - result not zero
@wtbs	inx			;2
	lda	waveTab,x	;4
	sta	$5011		;4
	ldy	sampleRate	;+3=13
-	dey			;2
	bne	-		;3
				;-1
	dec	samplePtr	;+5=22
	beq	@wtbl		;3
				;-1
	jmp	@wtbs		;3
@wtbe	rts
 .endif

; playWaveRate
;  Plays the wave table on loop, fixed-point rate approach.
;
;  For better waveform accuracy than the PSG's built-in waveforms.
;  A sine wave loader is implemented below, and used in this demo.
;
;  Parameters:
;   X:Y:samplePtr[0] is the sample position.
;   The rate is all 24 bits of sampleRate.
;
;  The output rate is 48,372+96/407; a wave rate of ~188.95404637592 Hz.

playWaveRate
	ldy	samplePtr+2
	ldx	samplePtr+3
	jmp	@wtqs

@wtqn	lda	waveTab,x	;4
	sta	$5011		;4
	lda	samplePtr+1	;3
	adc	sampleRate	;3
	sta	samplePtr+1	;3
	tya			;2
	adc	sampleRate+1	;3
	tay			;2
	txa			;2
	adc	sampleRate+2	;3
	tax			;2
@wtqs	lda	PCMEnable	;3
	bne	@wtqn		;+3=37

@wtqe	sty	samplePtr+2
	stx	samplePtr+3
	rts

; ==== main code

reset	sei		; disable interrupts asap
	lda	#$40
	sta	$4017	; disable APU frame IRQs
	lda	$4015	; clear any pending IRQs
	lda	#<(empty+63>>6)
	sta	$4012	; DMC sample address (8 bits raised to 15-bit)
	lda	#$84	; loop, rate #5 - (n*16+1)/782.24332486 sec
	sta	$4010	; DMC format
	lda	#0
	sta	$4013	; DMC count = 1 (IRQ immediately)
	sta	$5010	; enable MMC5 PCM output
	sta	$5100	; set banking to full 32KB
 .if 0
  ; clear low memory
	sta	samplePtr
	sta	samplePtr+1
	tay
-	sta	(samplePtr),y
	iny
	bne	-
 .endif
	jsr	gensine	; fill the waveTable
	; init. done; enable interrupts, and go
	cli
replay
;	lax	#0	; just under a second
;	ldy	#1	; highest
;	jsr	playWaveLenCtr

 .if defined(snd0)
	lda	#<snd0
	ldx	#>snd0
	ldy	#-1
	sty	PCMEnable
	jsr	playSound
 .endif
 .if defined(snd1)
	lda	#<snd1
	ldx	#>snd1
	ldy	#-1
	jsr	playSound
 .endif
	lda	#0
	sta	samplePtr+1
	sta	samplePtr+2
	sta	samplePtr+3
	lda	#32
	sta	sampleRate
	lda	#<669	; middle B (493.88234Hz, or MIDI note 58.999966)
	sta	sampleRate+1
	lda	#>669
	sta	sampleRate+2
	lda	#46	; DMC len = 783 (0.962616 secs)
	sta	PCMEnable
	lda	#$10
	sta	$4015
	jsr	playWaveRate
	lda	#0
	sta	sampleRate
	lda	#<128	; middle B (493.88234Hz, or MIDI note 58.999966)
	sta	sampleRate+1
	lda	#>128
	sta	sampleRate+2
	lda	#222	; DMC len = 783 (0.962616 secs)
	sta	PCMEnable
	lda	#$10
	sta	$4015
	jsr	playWaveRate
 .if defined(doscale)
	ldx	#0
@scalLp	lda	#12	; DMC len = 209 (0.2671803 secs)
	sta	PCMEnable
	lda	#$10
	sta	$4015
	stx	tmp+4
	ldy	scalTs,x
	lda	#<snd2
	ldx	#>snd2
	jsr	playSound
-	lda	PCMEnable
	bne	-

	ldx	tmp+4
	inx
	cpx	#scalTe-scalTs
	bne	@scalLp
 .endif
 .if defined(dompython)
	lda	#<mpython
	sta	tmp+4
	lda	#>mpython
	sta	tmp+5
@tunLp
	ldx	#8
	ldy	#0
	lda	(tmp+4),y
	beq	tunend
	bpl	isnote

	and	#127
	asl
	asl
	tax
	inc	tmp+4
	bne	+

	inc	tmp+5
+	lda	(tmp+4),y	;better be a note
isnote	inc	tmp+4
	bne	+

	inc	tmp+5
+	tay
	dex
	stx	PCMEnable
	lda	#$10
	sta	$4015
	lda	#<snd2
	ldx	#>snd2
	jsr	playSound
-	lda	PCMEnable
	bne	-
	beq	@tunLp
tunend
 .endif
	jmp	replay

 .if defined(doscale)
scalTs	dcb	48,50,52,53,55,57,59,60,62,64,65,67,69,71,72
scalTe
 .endif
 .if defined(dompython)
mpython
        dcb	72,$84,69,69,69,68,69,$84,77,72,$84,72,69,$84,70,70
        dcb	$84,70,72,$86,74,$84,74,70,$84,67,67,67,66,67,$84,76
        dcb	74,$84,74,70,$84,69,69,$84,69,70,$86,72,$84,72,74,$84
        dcb	69,69,69,68,69,$84,81,77,$84,77,72,$84,71,79,$84,79
        dcb	79,$86,79,$84,79,77,$84,76,79,79,78,79,$84,74,79,79
        dcb	78,79,$84,72,71,$84,72,71,$86,72,$86,72,$84,69,69,69
        dcb	68,69,$84,77,72,$84,72,69,$84,70,70,$84,70,72,$86,74
        dcb	$84,74,70,$84,67,67,67,66,67,$84,76,74,$84,74,70,$84
        dcb	69,69,$84,69,70,$86,72,$84,72,74,$84,69,69,69,68,69
        dcb	$84,81,77,$84,77,72,$84,71,79,$84,79,79,$86,79,$84,79
        dcb	77,$84,76,79,79,78,79,$84,74,79,79,78,79,$84,72,71
        dcb	$84,72,71,$86,72,$86,70,69,68,69,$84,74,72,$86,69,$86
        dcb	65,$86,62,$86,67,$86,65,$84,65,65,67,69,70,$84,76,74
        dcb	$86,72,$86,77,$86,76,$86,74,$86,72,$84,72,72,$84,74,$81,74
        dcb	$81,76,74,73,74,$86,76,$86,76,$84,77,$81,77,$81,81,79,77,79
	dcb	$86,81,$84,81;,81
	dcb	81,$84,79,77,$84,74,70,$86,69,$86,65,$86
        dcb	67,$86,64,$86,65,$84,65,72,69,68,69,$84,74,72,$86,69
        dcb	$86,65,$86,62,$86,67,$86,65,$84,65,65,67,69,70,$84,76
        dcb	74,$86,72,$86,77,$86,76,$86,74,$86,72,$84,72,72,$84,74
        dcb	$81,74,$81,76,74,73,74,$86,76,$86,76,$84,77,$81,77,$81,81,79,77,79
        dcb	$86,81,$84,81,81;,81
	dcb	$84,79,77,$84,74,70,$86,69,$86,65
        dcb	$86,67,$86,64,$86,65,$84,65,65,$86,74,$86,75,$86,76,$86
        dcb	77,$84,74,74,$84,74,73,$86,74,$84,74,65,$86,74,$86,75
        dcb	$86,76,$86,77,$84,75,75,$84,75,74,$86,75,$84,75,74,$86
        dcb	72,$86,71,$84,72,65,$84,64,65,$86,74,$86,73,$84,74,65
        dcb	$84,64,65,$86,69,$86,72,$84,70,72,$84,67,70,$84,69,70
        dcb	$84,67,69,$84,65,67,$84,63,65,$86,74,$86,75,$86,76,$86
        dcb	77,$84,74,74,$84,74,73,$86,74,$84,74,65,$86,74,$86,75
        dcb	$86,76,$86,77,$84,75,75,$84,75,74,$86,75,$84,75,74,$86
        dcb	72,$86,71,$84,72,65,$84,64,65,$86,74,$86,73,$84,74,65
        dcb	$84,64,65,$86,67,$86,75,$84,74,70,$84,72,69,$8A,70,$90
        dcb	70,75,75,74,72,$90,70,75,75,74,72,$90,70,75,75,74
        dcb	74,$84,70,70,79,79,77,75,$90,74,79,79,77,75,$90,74
        dcb	79,79,77,75,$84,74,74,76,74,76,$86,77,$86,77,$96,89
        dcb	76,$84,77,84,$90,89,76,$84,89,84,$86,89,$88,76,$8A,77
        dcb	$8A,78,$86,77,$86,75,$86,74,$86,75,$86,76,$86,77,$84,86
        dcb	86,$84,86,85,$86,86,$84,86,77,$86,74,$86,75,$86,76,$86
        dcb	77,$84,87,87,$84,87,86,$86,87,$84,87,86,$86,84,$86,83
        dcb	$84,84,77,$84,76,77,$86,86,$86,85,$84,86,77,$84,76,77
        dcb	$86,81,$86,84,$84,82,84,$84,79,82,$84,81,82,$84,79,81
        dcb	$84,77,79,$84,75,77,$86,74,$86,75,$86,76,$86,77,$84,86
        dcb	86,$84,86,85,$86,86,$84,86,77,$86,74,$86,75,$86,76,$86
        dcb	77,$84,87,87,$84,87,86,$86,87,$84,87,86,$86,84,$86,83
        dcb	$84,84,77,$84,76,77,$86,86,$86,85,$84,86,77,$84,76,77
        dcb	$86,79,$86,87,$84,86,82,$84,84,81,$8A,82,$90,82,75,75
        dcb	74,72,$90,70,75,75,74,72,$90,70,75,75,74,74,$84,70
        dcb	70,79,79,77,75,$90,74,79,79,77,75,$90,74,79,79,77
        dcb	75,$84,74,74,76,74,76,$86,77,$86,77,$96,89,76,$84,77
        dcb	84,$90,89,76,$84,89,84,$86,89,$88,76,$8A,77,$8A,78,$86
        dcb	77,$86,75,$86,74,$86,75,$86,76,$86,77,$84,86,86,$84,86
        dcb	85,$86,86,$84,86,77,$86,74,$86,75,$86,76,$86,77,$84,87
        dcb	87,$84,87,86,$86,87,$84,87,86,$86,84,$86,83,$84,84,77
        dcb	$84,76,77,$86,86,$86,85,$84,86,77,$84,76,77,$86,81,$86
        dcb	84,$84,82,84,$84,79,82,$84,81,82,$84,79,81,$84,77,79
        dcb	$84,75,77,$86,74,$86,75,$86,76,$86,77,$84,86,86,$84,86
        dcb	85,$86,86,$84,86,77,$86,74,$86,75,$86,76,$86,77,$84,87
        dcb	87,$84,87,86,$86,87,$84,87,86,$86,84,$86,83,$84,84,77
        dcb	$84,76,77,$86,86,$86,85,$84,86,77,$84,76,77,$86,79,$86
        dcb	87,$84,86,82,$84,84,81,$8A,82,$84,82,0
 .endif
; NMI handler
; should only be enabled when graphics are in use, and not playing raw PCM
nmi	rti

 ;NMI second 24-bit fraction (if were to ever develop that stop watch ROM...)
 ; 4.$427A NTSC
 ; 5.$1E89 PAL

; IRQ handler
; only used here as a silent DMC timer (yes, timer) for stopping PCM play
irq	pha
	inc	PCMEnable
	bne	+

	lda	#0
	sta	$4013
	sta	$4015
	pla
	rti

+	lda	PCMEnable
	sta	$4013
	lda	#-1
	sta	PCMEnable
	lda	#$10
	sta	$4015
	pla
	rti

empty

ORG	$FF80-24

 ;2**(x/12.+31)/srate
 ;divised this table so I no longer have to manually build on changes to the sample rate
ssrat		= srate>>srcale+2/5
noteTab dcw	2147483648/ssrat*125+(2147483648%ssrat*250/ssrat+1>>1)
	dcw	2275179671/ssrat*125+(2275179671%ssrat*250/ssrat+1>>1)
	dcw	2410468894/ssrat*125+(2410468894%ssrat*250/ssrat+1>>1)
	dcw	2553802834/ssrat*125+(2553802834%ssrat*250/ssrat+1>>1)
	dcw	2705659852/ssrat*125+(2705659852%ssrat*250/ssrat+1>>1)
	dcw	2866546760/ssrat*125+(2866546760%ssrat*250/ssrat+1>>1)
	dcw	3037000500/ssrat*125+(3037000500%ssrat*250/ssrat+1>>1)
	dcw	3217589947/ssrat*125+(3217589947%ssrat*250/ssrat+1>>1)
	dcw	3408917802/ssrat*125+(3408917802%ssrat*250/ssrat+1>>1)
	dcw	3611622603/ssrat*125+(3611622603%ssrat*250/ssrat+1>>1)
	dcw	3826380858/ssrat*125+(3826380858%ssrat*250/ssrat+1>>1)
	dcw	4053909305/ssrat*125+(4053909305%ssrat*250/ssrat+1>>1)

 .if 0
 ; for 40+2/256 cc (~44,735.5807Hz)
 ; 2**(x/12.)*24001.964558467758+.5
noteTab	dcw	24002,25429,26941,28543,30241,32039,33944,35962,38101,40366,42767,45310
 .endif

 ; Kind of amazing - I was able to pack all this in before the vector table
 ;on the first try - with one byte to spare.  Had to add the SEC for SBC, so zero to spare.
costab	dcb	32768/257,32758/257,32729/257,32679/257,32610/257,32522/257,32413/257,32286/257
	dcb	32138/257,31972/257,31786/257,31581/257,31357/257,31114/257,30853/257,30572/257
	dcb	30274/257,29957/257,29622/257,29269/257,28899/257,28511/257,28106/257,27684/257
	dcb	27246/257,26791/257,26320/257,25833/257,25330/257,24812/257,24279/257,23732/257
	dcb	23170/257,22595/257,22006/257,21403/257,20788/257,20160/257,19520/257,18868/257
	dcb	18205/257,17531/257,16846/257,16151/257,15447/257,14733/257,14010/257,13279/257
	dcb	12540/257,11793/257,11039/257,10279/257,9512/257,8740/257,7962/257,7180/257
	dcb	6393/257,5602/257,4808/257,4011/257,3212/257,2411/257,1608/257,804/257,0/257

gensine
	ldy	#0
	ldx	#gensine-costab-1
	clc
-	lda	#128
	adc	costab,x
	sta	waveTab,y
	iny
	dex
	bne	-

-	lda	#128
	adc	costab,x
	sta	waveTab,y
	inx
	iny
	bpl	-

	ldx	#gensine-costab-1
	sec
-	lda	#128
	sbc	costab,x
	sta	waveTab,y
	iny
	dex
	bne	-

-	lda	#128
	sbc	costab,x
	sta	waveTab,y
	inx
	iny
	bmi	-
	rts

ORG	$FFFA

	dcw	nmi
	dcw	reset
	dcw	irq
